import unittest

from hamcrest import *

from kerosene.configs.configs import ModelConfiguration
from kerosene.configs.exceptions import InvalidConfigurationError
from kerosene.configs.parsers import YamlConfigurationParser


class TestModelTrainerConfiguration(unittest.TestCase):
    VALID_CONFIG_FILE_PATH = "tests/configs/valid_config.yml"
    INVALID_CONFIG_FILE_PATH = "tests/configs/invalid_config.yml"

    MODELS_CONFIG_YML_TAG = "models"

    SIMPLE_NET_NAME = "SimpleNet"
    SIMPLE_NET_TYPE = "SimpleNet"

    SIMPLE_NET_OPTIMIZER_TYPE = "SGD"
    SIMPLE_NET_OPTIMIZER_PARAMS = {'lr': 0.01, 'momentum': 0.5, 'weight_decay': 0}

    SIMPLE_NET_SCHEDULER_TYPE = "ReduceLROnPlateau"
    SIMPLE_NET_SCHEDULER_PARAMS = {'mode': 'min', 'factor': 0.1, 'patience': 3}

    SIMPLE_NET_CRITERION_TYPE = "L1Loss"
    SIMPLE_NET_CRITERION_TYPE_2 = "MSELoss"

    SIMPLE_NET_METRIC_TYPE_1 = "Dice"
    SIMPLE_NET_METRIC_PARAMS_1 = {"num_classes": 4, "reduction": None, "ignore_index": 0, "average": None,
                                  "weight": None}

    SIMPLE_NET_METRIC_TYPE_2 = "Accuracy"

    SIMPLE_NET_GRADIENT_CLIPPING = {"clipping_strategy": "norm", "params": {"max_norm": 1.0}}

    def test_should_parse_valid_model_trainer_config(self):
        expected_config_dict = {self.SIMPLE_NET_NAME: {'type': self.SIMPLE_NET_TYPE,
                                                       'optimizer': {'type': self.SIMPLE_NET_OPTIMIZER_TYPE,
                                                                     'params': self.SIMPLE_NET_OPTIMIZER_PARAMS},
                                                       'scheduler': {'type': self.SIMPLE_NET_SCHEDULER_TYPE,
                                                                     'params': self.SIMPLE_NET_SCHEDULER_PARAMS},
                                                       'criterion': {"cycle": {'type': self.SIMPLE_NET_CRITERION_TYPE},
                                                                     "gan": {'type': self.SIMPLE_NET_CRITERION_TYPE_2}},
                                                       'metrics': {'Dice':{'type': self.SIMPLE_NET_METRIC_TYPE_1,
                                                                    'params': self.SIMPLE_NET_METRIC_PARAMS_1},
                                                                   'Accuracy': {'type': self.SIMPLE_NET_METRIC_TYPE_2}},
                                                       'gradients': self.SIMPLE_NET_GRADIENT_CLIPPING}}
        config_dict = YamlConfigurationParser.parse_section(self.VALID_CONFIG_FILE_PATH, self.MODELS_CONFIG_YML_TAG)
        model_trainer_config = ModelConfiguration.from_dict(self.SIMPLE_NET_NAME,
                                                            config_dict[self.SIMPLE_NET_NAME])

        assert_that(config_dict, equal_to(expected_config_dict))

        assert_that(model_trainer_config.optimizer_type, equal_to(self.SIMPLE_NET_OPTIMIZER_TYPE))
        assert_that(model_trainer_config.optimizer_params, equal_to(self.SIMPLE_NET_OPTIMIZER_PARAMS))
        assert_that(model_trainer_config.scheduler_type, equal_to(self.SIMPLE_NET_SCHEDULER_TYPE))
        assert_that(model_trainer_config.scheduler_params, equal_to(self.SIMPLE_NET_SCHEDULER_PARAMS))
        assert_that(model_trainer_config.criterions_configs[0].type, equal_to(self.SIMPLE_NET_CRITERION_TYPE))
        assert_that(model_trainer_config.criterions_configs[1].type, equal_to(self.SIMPLE_NET_CRITERION_TYPE_2))
        assert_that(model_trainer_config.metrics_configs[0].type, equal_to(self.SIMPLE_NET_METRIC_TYPE_1))
        assert_that(model_trainer_config.metrics_configs[0].params, equal_to(self.SIMPLE_NET_METRIC_PARAMS_1))

    def test_should_throw_on_invalid_model_trainer_config(self):
        config_dict = YamlConfigurationParser.parse_section(self.INVALID_CONFIG_FILE_PATH, self.MODELS_CONFIG_YML_TAG)

        assert_that(calling(ModelConfiguration.from_dict).with_args(self.SIMPLE_NET_NAME,
                                                                    config_dict[self.SIMPLE_NET_NAME]),
                    raises(InvalidConfigurationError))



# -*- coding: utf-8 -*-
# Copyright 2019 Kerosene Authors. All Rights Reserved.
#
# Licensed under the MIT License;
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://opensource.org/licenses/MIT
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from kerosene.configs.parsers import YamlConfigurationParser


class VisdomConfiguration(object):
    def __init__(self, port, server, env):
        self._port = port
        self._server = server
        self._env = env

    @property
    def port(self):
        return self._port

    @property
    def server(self):
        return self._server

    @property
    def env(self):
        return self._env

    @classmethod
    def from_dict(cls, config_dict):
        return cls(config_dict['port'], config_dict['server'], config_dict['env'])

    @classmethod
    def from_yml(cls, yml_file, yml_tag="visdom"):
        config = YamlConfigurationParser.parse_section(yml_file, yml_tag)
        return VisdomConfiguration.from_dict(config)



import pathlib

from setuptools import find_packages, setup

# The text of the README file
README_CONTENT = (pathlib.Path(__file__).parent / "README.md").read_text()

setup(
    name='torch-kerosene',
    version='0.2.2',
    description='Deep Learning framework for fast and clean research development with Pytorch',
    long_description=README_CONTENT,
    long_description_content_type='text/markdown',
    author='Benoit Anctil-Robitaille',
    author_email='benoit.anctil-robitaille.1@ens.etsmtl.ca',
    license='MIT',
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7"],
    packages=find_packages(exclude=("tests",)),
    install_requires=['numpy>=1.16.1',
                      'visdom>=0.1.8.8',
                      'pytorch-ignite>= 0.2.0',
                      'torch>=1.1',
                      'torchvision>=0.2.1',
                      'PyYAML']
)



import os
import shutil
import unittest
from unittest.mock import PropertyMock

import mock
import mockito
import torch
from hamcrest import *
from ignite.metrics import Accuracy
from torch import nn
from torch.optim import Optimizer, lr_scheduler

from kerosene.events import MonitorMode, Moment, Frequency, Phase
from kerosene.events.handlers.checkpoints import Checkpoint
from kerosene.nn.utils.gradients import GradientClippingStrategy
from kerosene.training.events import Event
from kerosene.training.trainers import SimpleTrainer, ModelTrainer


class ModelCheckpointIfBetterTest(unittest.TestCase):
    MODEL_NAME = "test_model"
    TRAINER_NAME = "test_trainer"
    SAVE_PATH = "tests/output/"

    def setUp(self) -> None:
        self._model_mock = mockito.mock(nn.Module)
        self._criterion_mock = {"BCE": mockito.mock(nn.CrossEntropyLoss)}
        self._optimizer_mock = mockito.mock(Optimizer)
        self._scheduler_mock = mockito.mock(lr_scheduler)
        self._metric_computer_mock = {"Accuracy": mockito.mock(Accuracy)}
        self._gradient_clipping_strategy = mockito.mock(GradientClippingStrategy)

        self._model_trainer = ModelTrainer(self.MODEL_NAME, self._model_mock, self._criterion_mock,
                                           self._optimizer_mock, self._scheduler_mock, self._metric_computer_mock,
                                           self._gradient_clipping_strategy)

        self._trainer_mock = mockito.mock(SimpleTrainer)

    def tearDown(self) -> None:
        if os.path.exists(self.SAVE_PATH):
            shutil.rmtree(self.SAVE_PATH)

    @mock.patch("kerosene.training.trainers.ModelTrainer.optimizer_state", new_callable=PropertyMock)
    @mock.patch("kerosene.training.trainers.ModelTrainer.model_state", new_callable=PropertyMock)
    @mock.patch("kerosene.training.trainers.ModelTrainer.valid_loss", new_callable=PropertyMock)
    def test_should_not_save_model_with_higher_valid_loss(self, valid_loss_mock, model_state_mock,
                                                          optimizer_state_mock):
        self._handler_mock = mockito.spy(
            Checkpoint(self.SAVE_PATH, lambda model_trainer: model_trainer.valid_loss, 0.01, MonitorMode.MIN))
        valid_loss_mock.return_value = torch.tensor([0.5])
        model_state_mock.return_value = dict
        optimizer_state_mock.return_value = dict
        self._trainer_mock.epoch = 5
        self._trainer_mock.model_trainers = [self._model_trainer]
        self._handler_mock(Event.ON_EPOCH_END(Moment(5, Frequency.EPOCH, Phase.TRAINING)), None, self._trainer_mock)
        valid_loss_mock.return_value = torch.tensor([0.6])
        self._handler_mock(Event.ON_EPOCH_END(Moment(5, Frequency.EPOCH, Phase.TRAINING)), None, self._trainer_mock)
        assert_that(bool(mockito.expect(self._handler_mock, times=1)._save_model(...).thenReturn(False)), is_(True))
        assert_that(
            not os.path.exists(os.path.join(self.SAVE_PATH, self.MODEL_NAME, self.MODEL_NAME + "_optimizer.tar")))

    @mock.patch("kerosene.training.trainers.ModelTrainer.optimizer_state", new_callable=PropertyMock)
    @mock.patch("kerosene.training.trainers.ModelTrainer.model_state", new_callable=PropertyMock)
    @mock.patch("kerosene.training.trainers.ModelTrainer.valid_loss", new_callable=PropertyMock)
    def test_should_save_model_with_lower_valid_loss(self, valid_loss_mock, model_state_mock, optimizer_state_mock):
        self._handler_mock = mockito.spy(
            Checkpoint(self.SAVE_PATH, lambda model_trainer: model_trainer.valid_loss, 0.01, MonitorMode.MIN))
        valid_loss_mock.return_value = torch.tensor([0.5])
        model_state_mock.return_value = dict
        optimizer_state_mock.return_value = dict
        self._trainer_mock.epoch = 5
        self._trainer_mock.model_trainers = [self._model_trainer]
        self._handler_mock(Event.ON_EPOCH_END(Moment(5, Frequency.EPOCH, Phase.TRAINING)), None, self._trainer_mock)
        valid_loss_mock.return_value = torch.tensor([0.4])
        self._handler_mock(Event.ON_EPOCH_END(Moment(5, Frequency.EPOCH, Phase.TRAINING)), None, self._trainer_mock)
        assert_that(bool(mockito.expect(self._handler_mock, times=1)._save_model(...).thenReturn(True)), is_(True))
        assert_that(os.path.exists(os.path.join(self.SAVE_PATH, self.MODEL_NAME, self.MODEL_NAME + ".tar")))

    @mock.patch("kerosene.training.trainers.ModelTrainer.optimizer_state", new_callable=PropertyMock)
    @mock.patch("kerosene.training.trainers.ModelTrainer.model_state", new_callable=PropertyMock)
    @mock.patch("kerosene.training.trainers.ModelTrainer.valid_loss", new_callable=PropertyMock)
    def test_should_save_optimizer_with_lower_valid_loss(self, valid_loss_mock, model_state_mock, optimizer_state_mock):
        self._handler_mock = mockito.spy(
            Checkpoint(self.SAVE_PATH, lambda model_trainer: model_trainer.valid_loss, 0.01, MonitorMode.MIN))
        valid_loss_mock.return_value = torch.tensor([0.5])
        model_state_mock.return_value = dict
        optimizer_state_mock.return_value = dict
        self._trainer_mock.model_trainers = [self._model_trainer]
        self._handler_mock(Event.ON_EPOCH_END(Moment(5, Frequency.EPOCH, Phase.TRAINING)), None, self._trainer_mock)
        valid_loss_mock.return_value = torch.tensor([0.4])
        self._handler_mock(Event.ON_EPOCH_END(Moment(5, Frequency.EPOCH, Phase.TRAINING)), None, self._trainer_mock)
        assert_that(bool(mockito.expect(self._handler_mock, times=1)._save_model(...).thenReturn(True)), is_(True))
        assert_that(
            os.path.exists(os.path.join(self.SAVE_PATH, self.MODEL_NAME, self.MODEL_NAME + "_optimizer.tar")))

    @mock.patch("kerosene.training.trainers.ModelTrainer.optimizer_state", new_callable=PropertyMock)
    @mock.patch("kerosene.training.trainers.ModelTrainer.model_state", new_callable=PropertyMock)
    @mock.patch("kerosene.training.trainers.ModelTrainer.valid_metrics", new_callable=PropertyMock)
    def test_should_save_optimizer_with_higher_valid_metric(self, valid_metrics_mock, model_state_mock,
                                                            optimizer_state_mock):
        self._handler_mock = mockito.spy(
            Checkpoint(self.SAVE_PATH, lambda model_trainer: model_trainer.valid_metrics, 0.01, MonitorMode.MAX))
        valid_metrics_mock.return_value = torch.tensor([0.5])
        model_state_mock.return_value = dict
        optimizer_state_mock.return_value = dict
        self._trainer_mock.model_trainers = [self._model_trainer]
        self._handler_mock(Event.ON_EPOCH_END(Moment(5, Frequency.EPOCH, Phase.TRAINING)), None, self._trainer_mock)
        valid_metrics_mock.return_value = torch.tensor([0.8])
        self._handler_mock(Event.ON_EPOCH_END(Moment(5, Frequency.EPOCH, Phase.TRAINING)), None, self._trainer_mock)
        assert_that(bool(mockito.expect(self._handler_mock, times=1)._save_optimizer(...).thenReturn(True)), is_(True))
        assert_that(
            os.path.exists(os.path.join(self.SAVE_PATH, self.MODEL_NAME, self.MODEL_NAME + "_optimizer.tar")))

    @mock.patch("kerosene.training.trainers.ModelTrainer.optimizer_state", new_callable=PropertyMock)
    @mock.patch("kerosene.training.trainers.ModelTrainer.model_state", new_callable=PropertyMock)
    @mock.patch("kerosene.training.trainers.ModelTrainer.valid_metrics", new_callable=PropertyMock)
    def test_should_not_save_optimizer_with_lower_valid_metric(self, valid_metrics_mock, model_state_mock,
                                                               optimizer_state_mock):
        self._handler_mock = mockito.spy(
            Checkpoint(self.SAVE_PATH, lambda model_trainer: model_trainer.valid_metrics, 0.01, MonitorMode.MAX))

        valid_metrics_mock.return_value = torch.tensor([0.8])
        model_state_mock.return_value = dict
        optimizer_state_mock.return_value = dict
        self._trainer_mock.model_trainers = [self._model_trainer]
        self._handler_mock(Event.ON_EPOCH_END(Moment(5, Frequency.EPOCH, Phase.TRAINING)), None, self._trainer_mock)
        valid_metrics_mock.return_value = torch.tensor([0.5])
        self._handler_mock(Event.ON_EPOCH_END(Moment(5, Frequency.EPOCH, Phase.TRAINING)), None, self._trainer_mock)
        assert_that(bool(mockito.expect(self._handler_mock, times=1)._save_model(...).thenReturn(False)), is_(True))
        assert_that(
            not os.path.exists(os.path.join(self.SAVE_PATH, self.MODEL_NAME, self.MODEL_NAME + "_optimizer.tar")))



import logging

from kerosene.events import MonitorMode, TemporalEvent
from kerosene.events.handlers.base_monitor_watcher import MonitorWatcher, MonitorPatienceExceeded
from kerosene.training.trainers import Trainer


class EarlyStopping(MonitorWatcher):
    LOGGER = logging.getLogger("EarlyStopping")

    def __init__(self, monitor_fn, mode: MonitorMode, min_delta=0.01, patience=3):
        super(EarlyStopping, self).__init__(monitor_fn, mode, min_delta, patience)

    def __call__(self, temporal_event: TemporalEvent, monitors, trainer: Trainer):
        for model_trainer in trainer.model_trainers:
            try:
                value = self._monitor_fn(model_trainer)
                self.watch(model_trainer.name, value)
            except MonitorPatienceExceeded as e:
                model_trainer.finalize()



from typing import Union, List


class VisdomData(object):

    def __init__(self, source_name, variable_name: Union[List[str], str], plot_type, plot_frequency, x, y,
                 params: dict = None):
        self._source_name = source_name
        self._variable_name = variable_name
        self._plot_type = plot_type
        self._plot_frequency = plot_frequency
        self._x = x
        self._y = y
        self._params = params

    @property
    def id(self):
        return hash(self)

    @property
    def source_name(self):
        return self._source_name

    @property
    def variable_name(self):
        return self._variable_name

    @property
    def plot_type(self):
        return self._plot_type

    @property
    def plot_frequency(self):
        return self._plot_frequency

    @property
    def x(self):
        return self._x

    @property
    def y(self):
        return self._y

    @property
    def params(self):
        return self._params

    def __hash__(self):
        return hash(self._source_name + self._variable_name + str(self._plot_frequency) + str(self._plot_type))

    def __str__(self):
        return "Source: {} Variable: {} Plot Type: {} Frequency: {} x: {} y: {}".format(self.source_name,
                                                                                        self.variable_name,
                                                                                        str(self.plot_type),
                                                                                        str(self.plot_frequency),
                                                                                        self.x, self.y)



import unittest

import torch
from hamcrest import assert_that, equal_to

from kerosene.nn.apex import ApexLoss


class TestApexLoss(unittest.TestCase):
    VALID_LOSS_VALUE = torch.tensor([1.0, 2.0, 3.0])
    INVALID_LOSS_VALUE = "Ford Fiesta"

    VALID_SCALAR = 2.0
    INVALID_SCALAR = "Ford Focus"

    VALID_LOSS_ID = 0
    INVALID_LOSS_ID = "Ford F150"

    def test_should_add_losses(self):
        expected_result = ApexLoss(self.VALID_LOSS_ID, torch.tensor([2.0, 4.0, 6.0]), None)

        apex_loss1 = ApexLoss(self.VALID_LOSS_ID, self.VALID_LOSS_VALUE, None)
        apex_loss2 = ApexLoss(self.VALID_LOSS_ID, self.VALID_LOSS_VALUE, None)

        assert_that(apex_loss1 + apex_loss2, equal_to(expected_result))

    def test_should_multiply_by_a_scalar(self):
        expected_result = ApexLoss(self.VALID_LOSS_ID, self.VALID_SCALAR * self.VALID_LOSS_VALUE, None)

        loss = ApexLoss(self.VALID_LOSS_ID, self.VALID_LOSS_VALUE, None)

        assert_that(self.VALID_SCALAR * loss, equal_to(expected_result))
        assert_that(loss * self.VALID_SCALAR, equal_to(expected_result))

    # noinspection PyTypeChecker
    def test_should_divide_by_a_scalar(self):
        left_div_expected_result = ApexLoss(self.VALID_LOSS_ID, self.VALID_LOSS_VALUE / self.VALID_SCALAR,
                                            None)
        right__div_expected_result = ApexLoss(self.VALID_LOSS_ID, self.VALID_SCALAR / self.VALID_LOSS_VALUE,
                                              None)

        loss = ApexLoss(self.VALID_LOSS_ID, self.VALID_LOSS_VALUE, None)

        assert_that(loss / self.VALID_SCALAR, equal_to(left_div_expected_result))
        assert_that(self.VALID_SCALAR / loss, equal_to(right__div_expected_result))

    def test_should_compute_the_mean(self):
        expected_result = ApexLoss(self.VALID_LOSS_ID, torch.tensor([2.0]), None)

        loss = ApexLoss(self.VALID_LOSS_ID, self.VALID_LOSS_VALUE, None)

        assert_that(loss.mean(), equal_to(expected_result))

    # noinspection PyTypeChecker
    def test_should_substract_a_scalar(self):
        left_sub_expected_result = ApexLoss(self.VALID_LOSS_ID, self.VALID_LOSS_VALUE - self.VALID_SCALAR, None)
        right_sub_expected_result = ApexLoss(self.VALID_LOSS_ID, self.VALID_SCALAR - self.VALID_LOSS_VALUE, None)

        loss = ApexLoss(self.VALID_LOSS_ID, self.VALID_LOSS_VALUE, None)

        assert_that(loss - self.VALID_SCALAR, equal_to(left_sub_expected_result))
        assert_that(self.VALID_SCALAR - loss, equal_to(right_sub_expected_result))



# -*- coding: utf-8 -*-
# Copyright 2019 Kerosene Authors. All Rights Reserved.
#
# Licensed under the MIT License;
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://opensource.org/licenses/MIT
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import abc
import os
from socket import socket
from typing import List

import torch

try:
    from torch.distributed import is_nccl_available

    NCCL_AVAILABLE = is_nccl_available()
except ImportError:
    NCCL_AVAILABLE = False

from kerosene.configs.exceptions import InvalidConfigurationError
from kerosene.utils.devices import get_devices, on_multiple_gpus, num_gpus, on_cpu


class HtmlConfiguration(object):

    @abc.abstractmethod
    def to_html(self):
        raise NotImplementedError()


class DatasetConfiguration(HtmlConfiguration):
    def __init__(self, config_dict):
        for key in config_dict:
            setattr(self, key, config_dict[key])

    def to_html(self):
        configuration_values = '\n'.join("<p>%s: %s</p>" % item for item in vars(self).items())
        return "<h2>Dataset Configuration</h2> \n {}".format(configuration_values)


class Configuration(object):
    def __init__(self, name, type, params):
        self._name = name
        self._type = type
        self._params = params

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def type(self):
        return self._type

    @type.setter
    def type(self, value):
        self._type = value

    @property
    def params(self):
        return self._params

    @params.setter
    def params(self, value):
        self._params = value


class ModelConfiguration(Configuration, HtmlConfiguration):
    def __init__(self, model_name, model_type, model_params, optimizer_type, optimizer_params, scheduler_type,
                 scheduler_params, criterions_configs: List[Configuration], metrics_configs: List[Configuration],
                 gradient_clipping_strategy, gradient_clipping_params):
        super().__init__(model_name, model_type, model_params)
        self._optimizer_type = optimizer_type
        self._optimizer_params = optimizer_params

        self._scheduler_type = scheduler_type
        self._scheduler_params = scheduler_params

        self._criterions_configs = criterions_configs
        self._metrics_configs = metrics_configs

        self._gradient_clipping_strategy = gradient_clipping_strategy
        self._gradient_clipping_params = gradient_clipping_params

    @property
    def optimizer_type(self):
        return self._optimizer_type

    @property
    def optimizer_params(self):
        return self._optimizer_params

    @property
    def scheduler_type(self):
        return self._scheduler_type

    @property
    def scheduler_params(self):
        return self._scheduler_params

    @property
    def criterions_configs(self):
        return self._criterions_configs

    @property
    def metrics_configs(self):
        return self._metrics_configs

    @property
    def gradient_clipping_strategy(self):
        return self._gradient_clipping_strategy

    @property
    def gradient_clipping_params(self):
        return self._gradient_clipping_params

    @classmethod
    def from_dict(cls, model_name, config_dict):
        try:
            return cls(model_name, config_dict["type"], config_dict.get("params"),
                       config_dict["optimizer"]["type"], config_dict["optimizer"].get("params"),
                       config_dict["scheduler"]["type"], config_dict["scheduler"].get("params"),
                       list(map(lambda criterion: Configuration(criterion, config_dict["criterion"][criterion]["type"],
                                                                config_dict["criterion"][criterion].get("params")),
                                config_dict["criterion"].keys())),
                       list(map(lambda metric: Configuration(metric, config_dict["metrics"][metric]["type"],
                                                             config_dict["metrics"][metric].get("params")),
                                config_dict["metrics"].keys())),
                       config_dict.get("gradients", {}).get("clipping_strategy"),
                       config_dict.get("gradients", {}).get("params"))
        except KeyError as e:
            raise InvalidConfigurationError(
                "The provided model configuration is invalid. The section {} is missing.".format(e))

    def to_html(self):
        configuration_values = '\n'.join("<p>%s: %s</p>" % item for item in vars(self).items())
        return "<h2>Model Configuration \n </h2><h4>{}</h4> \n {}".format(self.name, configuration_values)


class RunConfiguration(HtmlConfiguration):

    def __init__(self, use_amp: bool = True, amp_opt_level: str = 'O1', local_rank: int = 0,
                 world_size: int = num_gpus()):
        self._use_amp = use_amp
        self._amp_opt_level = amp_opt_level
        self._local_rank = local_rank
        self._world_size = world_size

        self._devices = get_devices()
        self._device = self._devices[self._local_rank]

        if not on_cpu(self._device):
            torch.cuda.set_device(self._device)
            self._initialize_ddp_process_group()

    @property
    def use_amp(self):
        return self._use_amp

    @property
    def amp_opt_level(self):
        return self._amp_opt_level

    @property
    def local_rank(self):
        return self._local_rank

    @property
    def world_size(self):
        return self._world_size

    @property
    def devices(self):
        return self._devices

    @property
    def device(self):
        return self._device

    @device.setter
    def device(self, device):
        self._device = device

    @staticmethod
    def _get_random_free_port():
        with socket() as s:
            s.bind(("", 0))
            return s.getsockname()[1]

    def _initialize_ddp_process_group(self):
        if on_multiple_gpus(self._devices):
            if NCCL_AVAILABLE:
                if os.environ.get("MASTER_ADDR") is None:
                    os.environ["MASTER_ADDR"] = "127.0.0.1"
                if os.environ.get("MASTER_PORT") is None:
                    os.environ["MASTER_PORT"] = str(self._get_random_free_port())
                if os.environ.get("WORLD_SIZE") is None:
                    os.environ["WORLD_SIZE"] = str(self._world_size)
                torch.distributed.init_process_group(backend='nccl', init_method='env://',
                                                     world_size=int(os.environ["WORLD_SIZE"]), rank=self._local_rank)
            else:
                raise Exception("NCCL not available and required for multi-GPU training.")

    def to_html(self):
        configuration_values = '\n'.join("<p>%s: %s</p>" % item for item in vars(self).items())
        return "<h2>Run Configuration</h2> \n {}".format(configuration_values)


class TrainerConfiguration(HtmlConfiguration):
    def __init__(self, config_dict):
        for key in config_dict:
            setattr(self, key, config_dict[key])

    def to_html(self):
        configuration_values = '\n'.join("<p>%s: %s</p>" % item for item in vars(self).items())
        return "<h2>Training Configuration</h2> \n {}".format(configuration_values)



# -*- coding: utf-8 -*-
# Copyright 2019 Kerosene Authors. All Rights Reserved.
#
# Licensed under the MIT License;
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://opensource.org/licenses/MIT
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import List

import torch


def on_cpu(device: torch.device):
    return str(device) == "cpu"


def on_gpu(device: torch.device):
    return device.type == "cuda"


def on_gpus(devices: List[torch.device]):
    return all([device.type == "cuda" for device in devices])


def on_single_device(devices: List[torch.device]):
    return len(devices) == 1


def on_multiple_devices(devices: List[torch.device]):
    return len(devices) > 1


def on_single_gpu(devices: List[torch.device]):
    return on_single_device(devices) and on_gpus(devices)


def on_multiple_gpus(devices: List[torch.device]):
    return on_multiple_devices(devices) and on_gpus(devices)


def get_devices():
    return [torch.device("cuda:{}".format(device_id)) for device_id in
            range(torch.cuda.device_count())] if torch.cuda.is_available() else [torch.device("cpu")]


def num_gpus():
    return len(list(([device.type == "cuda" for device in get_devices()])))



# -*- coding: utf-8 -*-
# Copyright 2019 Kerosene Authors. All Rights Reserved.
#
# Licensed under the MIT License;
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://opensource.org/licenses/MIT
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from enum import Enum
from typing import Union

from torch.optim import Optimizer
from torch.optim.lr_scheduler import ReduceLROnPlateau, StepLR, MultiStepLR, ExponentialLR, CosineAnnealingLR, \
    _LRScheduler


class SchedulerType(Enum):
    ReduceLROnPlateau = "ReduceLROnPlateau"
    StepLR = "StepLR"
    MultiStepLR = "MultiStepLR"
    ExponentialLR = "ExponentialLR"
    CosineAnnealingLR = "CosineAnnealingLR"
    CyclicLR = "CyclicLR"
    CosineAnnealingWarmRestarts = "CosineAnnealingWarmRestarts"


class SchedulerFactory(object):

    def __init__(self):
        self._schedulers = {
            "ReduceLROnPlateau": ReduceLROnPlateau,
            "StepLR": StepLR,
            "MultiStepLR": MultiStepLR,
            "ExponentialLR": ExponentialLR,
            "CosineAnnealingLR": CosineAnnealingLR,
        }

    def create(self, scheduler_type: Union[str, SchedulerType], optimizer: Optimizer, params):
        return self._schedulers[str(scheduler_type)](optimizer=optimizer, **params) if params is not None else \
            self._schedulers[str(scheduler_type)](optimizer=optimizer)

    def register(self, function: str, creator: _LRScheduler):
        """
        Add a new activation layer.
        Args:
           function (str): Activation layer name.
           creator (:obj:`torch.nn.Module`): A torch module object wrapping the new custom optimizer function.
        """
        self._schedulers[function] = creator
